
//			PALEDIT command, and related functions and structures:

#ifndef PALEDIT_VERSION
#define PALEDIT_VERSION					0.99.7

#ifndef MACRO_STR
#define MACRO_EXP( arg )				#arg
#define MACRO_STR( arg )				MACRO_EXP( arg )
#endif	//	MACRO_STR

#include "resource.h"


#define ALT_BIT							0x80000
#define SPACE							0x20
#define ESC								0x1b



const wchar_t PalEditHelpText[]	= L"Edit the console palette interactively.\n\
\n\
PALEDIT /D /I filename \r\n\
\n\
\t/D\tload a built-in set of default colors\n\
\t/I\tallow incomplete palette files\n\
\n";


const wchar_t PalEditOpenFilter[]	= L"All files\0*.*\0Text files\0*.palette;*.txt\0INI files\0*.ini\0XML files\0*.itermcolors;*.xml\0";


const wchar_t PalEditRegKey[]		= L"Software\\JPPlugins\\Palette";

const wchar_t PalEditFlagsValue[]	= L"SFI dialog";


int CtrlBreakFlag				= 0;



struct _PalEditState {
	_PalEditState *Prev;
	_PalEditState *Next;
	unsigned int Changes;
	COLORREF Colors[16];
};



BOOL WINAPI CtrlBreakHandler( DWORD Type )
{
	if ( ( Type == CTRL_C_EVENT ) || ( Type == CTRL_BREAK_EVENT ) ) {
		CtrlBreakFlag = -1;
		return TRUE;
	}

	return FALSE;
}



WORD HighlightAttr()
{
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	WORD rv						= 0;
	int fg = 0, bg = 0;
	WORD hl						= 0;

	BYTE FirstColor[16]			= { 0x0f, 0x1f, 0x2f, 0x3f, 0x4f, 0x5f, 0x6f, 0x70, 0x8f, 0x9f, 0xa0, 0xb0, 0xcf, 0xd0, 0xe0, 0xf2 };
	BYTE SecondColor[16]		= { 0x0e, 0x1b, 0x2b, 0x3e, 0x4b, 0x5b, 0x6b, 0x79, 0x8b, 0x9b, 0xa9, 0xb9, 0xc0, 0xde, 0xec, 0xf0 };

	memset( &ConsoleInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFO ) );
	GetConsoleScreenBufferInfo( ConsoleHandle, &ConsoleInfo );
	rv = ConsoleInfo.wAttributes;

	fg = ( rv & 0x0f );
	bg = ( rv & 0xf0 ) >> 4;

	if ( fg == SecondColor[bg] )
		hl = FirstColor[bg];
	else
		hl = SecondColor[bg];

	SetConsoleTextAttribute( ConsoleHandle, hl );
	return rv;
}



int PalEdit_GetKey()
{
	int key			= 0;

	do {
		if ( CtrlBreakFlag ) {
			WORD Attr = HighlightAttr();
			_cputws( L"\r ^C\r\n" );
			SetConsoleTextAttribute( ConsoleHandle, Attr );
			return ESC;
		}

		if ( _kbhit() ) {
			key = _getch();

			if ( ( key == 0x00 ) || ( key == 0xe0 ) )
				key = _getch() | FBIT;

			if ( ( key >= 'a' ) && ( key <= 'z' ) )
				key = towupper( key );

			if ( ( key & FBIT ) == 0 )
				if ( GetAsyncKeyState( VK_MENU ) & 0x8000 )
					key |= ALT_BIT;

			return key;
		}
		else
			Sleep( 10 );
	}
	while ( TRUE );

	return key;
}



INT_PTR CALLBACK PalEdit_SaveInfoDialogProc( HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam )
{
	static _SaveInfo *SaveInfo		= NULL;


	switch ( Msg ) {

	case WM_INITDIALOG :

		SaveInfo = (_SaveInfo*) lParam;
			
		PostMessage( GetDlgItem( hWnd, IDC_EDIT1 ), EM_SETLIMITTEXT, SHORT_BUF_LEN - 1, 0 );
		PostMessage( GetDlgItem( hWnd, IDC_EDIT2 ), EM_SETLIMITTEXT, SHORT_BUF_LEN - 1, 0 );
		PostMessage( GetDlgItem( hWnd, IDC_EDIT3 ), EM_SETLIMITTEXT, SHORT_BUF_LEN - 1, 0 );
		PostMessage( GetDlgItem( hWnd, IDC_EDIT4 ), EM_SETLIMITTEXT, SHORT_BUF_LEN - 1, 0 );

		if ( SaveInfo ) {
			if ( SaveInfo->Name )
				SetDlgItemText( hWnd, IDC_EDIT1, SaveInfo->Name );

			if ( SaveInfo->Author )
				SetDlgItemText( hWnd, IDC_EDIT2, SaveInfo->Author );

			if ( SaveInfo->Url )
				SetDlgItemText( hWnd, IDC_EDIT3, SaveInfo->Url );

			if ( SaveInfo->License )
				SetDlgItemText( hWnd, IDC_EDIT4, SaveInfo->License );
		}

		if ( SaveInfo->DlgFlags > 2 )
			SaveInfo->DlgFlags = 1;

		CheckRadioButton( hWnd, IDC_RADIO1, IDC_RADIO3, IDC_RADIO1 + SaveInfo->DlgFlags );
		CenterWindow2( hWnd );

		return TRUE;

	case WM_COMMAND :

		switch ( LOWORD( wParam ) ) {

		case IDCANCEL :
		
			EndDialog( hWnd, IDCANCEL );
			return TRUE;

		case IDOK :

			if ( SaveInfo ) {
				if ( SaveInfo->Name )
					GetDlgItemText( hWnd, IDC_EDIT1, SaveInfo->Name, SHORT_BUF_LEN );

				if ( SaveInfo->Author )
					GetDlgItemText( hWnd, IDC_EDIT2, SaveInfo->Author, SHORT_BUF_LEN );

				if ( SaveInfo->Url )
					GetDlgItemText( hWnd, IDC_EDIT3, SaveInfo->Url, SHORT_BUF_LEN );

				if ( SaveInfo->License )
					GetDlgItemText( hWnd, IDC_EDIT4, SaveInfo->License, SHORT_BUF_LEN );

				if ( IsDlgButtonChecked( hWnd, IDC_RADIO1 ) )
					SaveInfo->DlgFlags = 0;
				else if ( IsDlgButtonChecked( hWnd, IDC_RADIO2 ) )
					SaveInfo->DlgFlags = 1;
				else
					SaveInfo->DlgFlags = 2;
			}

			EndDialog( hWnd, IDOK );
			return TRUE;

		default :

			return FALSE;
		}

	default :

		return FALSE;
	}
}



int PalEdit_SaveDlgFlags( unsigned int Value )
{
	HKEY KeyHandle		= NULL;
	DWORD value			= Value;


	if ( RegCreateKeyEx( HKEY_CURRENT_USER, PalEditRegKey, 0, NULL, 0, KEY_READ | KEY_WRITE, NULL, &KeyHandle, NULL ) == ERROR_SUCCESS ) {
		RegSetValueEx( KeyHandle, PalEditFlagsValue, 0, REG_DWORD, (BYTE*) &value, sizeof( DWORD ) );
		RegCloseKey( KeyHandle );
	}

	return 0;
}



unsigned int PalEdit_GetDlgFlags()
{
	HKEY KeyHandle		= NULL;

	DWORD DlgMode		= 1;
	DWORD ValueSize		= sizeof( DWORD );
	DWORD ValueType		= 0;


	if ( RegOpenKeyEx( HKEY_CURRENT_USER, PalEditRegKey, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {
		if ( RegQueryValueEx( KeyHandle, PalEditFlagsValue, NULL, &ValueType, (BYTE*) &DlgMode, &ValueSize ) != ERROR_SUCCESS )
			DlgMode = 1;

		if ( ( ValueType != REG_DWORD ) || ( ValueSize != sizeof( DWORD ) ) || ( DlgMode > 2 ) )
			DlgMode = 1;

		RegCloseKey( KeyHandle );
	}

	return DlgMode;
}



int PalEdit_ShowSaveInfoDlgMaybe( _SaveInfo *SaveInfo )
{
	if ( !SaveInfo )
		return -1;


	BOOL ShowDialog			= TRUE;

	int rv			= 0;


	if ( SaveInfo->DlgFlags == 1 ) {
	
		if ( SaveInfo->Name )
			if ( SaveInfo->Name[0] )
				ShowDialog = FALSE;

		if ( SaveInfo->Author )
			if ( SaveInfo->Author[0] )
				ShowDialog = FALSE;

		if ( SaveInfo->Url )
			if ( SaveInfo->Url[0] )
				ShowDialog = FALSE;

		if ( SaveInfo->License )
			if ( SaveInfo->License[0] )
				ShowDialog = FALSE;
	}
	else if ( SaveInfo->DlgFlags == 2 )
		ShowDialog = FALSE;

	if ( ShowDialog ) {
		rv = (int) DialogBoxParam( MyDLLHandle, MAKEINTRESOURCE( IDD_DIALOG1 ), MyWindow(), PalEdit_SaveInfoDialogProc, (LPARAM) SaveInfo );
		PalEdit_SaveDlgFlags( SaveInfo->DlgFlags );
	}
	
	return rv;
}



int PalEdit_UpdateStdColors( unsigned int bg, unsigned int fg )
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";


	wsprintf( temp, L"option //stdcolors=%s on %s", TccColorNames[fg], TccColorNames[bg] );
	return Command( temp, 0 );
}



int RgbLum( COLORREF Color )
{
	return ( ( 299 * GetRValue( Color ) ) + ( 587 * GetGValue( Color ) ) + ( 114 * GetBValue( Color ) ) ) / 255;
}



int PalEdit_ContrastyColor( int Index, COLORREF *Colors )
{
	int lum				= 0;
	int Darkest			= 0;
	int DarkestLum		= 1001;
	int Brightest		= 0;
	int BrightestLum	= -1;
	int IndexColorLum	= 0;
	int rv				= 0;


	for ( int i = 0; i < 16; i++ ) {
		lum = RgbLum( Colors[i] );

		if ( i == Index )
			IndexColorLum = lum;

		if ( lum < DarkestLum )
			Darkest = i, DarkestLum = lum;

		if ( lum > BrightestLum )
			Brightest = i, BrightestLum = lum;
	}

	if ( IndexColorLum >= 500 )
		rv = Darkest;
	else
		rv = Brightest;

	if ( rv == Index )
		rv ^= 8;

	return rv;
}



int PalEdit_UpdatePalette( COLORREF *Colors, unsigned int Index, CONSOLE_SCREEN_BUFFER_INFOEX *BufferInfo )
{
	if ( !Colors || !BufferInfo )
		return -1;


	CONSOLE_SCREEN_BUFFER_INFOEX BufInfo;


	memset( &BufInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFOEX ) );
	BufInfo.cbSize = sizeof( CONSOLE_SCREEN_BUFFER_INFOEX );


	if ( GetConsoleScreenBufferInfoEx( ConsoleHandle, &BufInfo ) ) {

		if ( Index < 16 )
			BufInfo.ColorTable[Index] = *Colors;
		else
			for ( int i = 0; i < 16; i++ )
				BufInfo.ColorTable[i] = Colors[i];

		BufInfo.srWindow.Right	+= 1;
		BufInfo.srWindow.Bottom	+= 1;

		SetConsoleScreenBufferInfoEx( ConsoleHandle, &BufInfo );
		RefreshWindows();
		memcpy( BufferInfo, &BufInfo, sizeof( CONSOLE_SCREEN_BUFFER_INFOEX ) );
		return 0;
	}
	else
		return 1;
}



COLORREF PalEdit_ColorPicker( unsigned int Index, CONSOLE_SCREEN_BUFFER_INFOEX *BufferInfo )
{
	wchar_t Title[SHORT_BUF_LEN]			= L"";

	CHOOSECOLOR ChooseColorData;

	COLORREF Colors[16];
	COLORREF InitColor			= 0;


	memset( &ChooseColorData, 0x00, sizeof( CHOOSECOLOR ) );
	ChooseColorData.lStructSize		= sizeof( CHOOSECOLOR );
	ChooseColorData.hwndOwner		= MyWindow();
	ChooseColorData.lpCustColors	= Colors;
	ChooseColorData.Flags			= CC_ANYCOLOR | CC_FULLOPEN | CC_RGBINIT | CC_ENABLEHOOK;
	ChooseColorData.lCustData		= (LPARAM) Title;
	ChooseColorData.lpfnHook		= PickColorProc;
	ChooseColorData.rgbResult		= BufferInfo->ColorTable[( Index < 16 ) ? Index : 0];

	for ( int i = 0; i < 16; i++ )
		Colors[i] = BufferInfo->ColorTable[i];

	wsprintf( Title, L"Edit color #%u - %s", Index, TccColorNames[Index] );
	
	if ( ChooseColor( &ChooseColorData ) )
		return ChooseColorData.rgbResult;
	else
		return -1;
}



int PalEdit_EraseTopLine()
{
	COORD Origin		= { 0, 0 };
	DWORD CharsWritten	= 0;


	FillConsoleOutputCharacter( ConsoleHandle, SPACE, 80, Origin, &CharsWritten );
	return 0;
}



int PalEdit_ShowUndoInfo( unsigned int Changes, unsigned int Latest )
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";

	COORD Origin		= { 0, 0 };
	DWORD CharsWritten	= 0;


	PalEdit_EraseTopLine();
	wsprintf( temp, L"%u / %u", Changes, Latest - Changes );
	Origin.X = 79 - (SHORT) wcslen( temp );
	WriteConsoleOutputCharacter( ConsoleHandle, temp, (DWORD) wcslen( temp ), Origin, &CharsWritten );
	return 0;
}



int PalEdit_ShowStringTop( LPTSTR Text )
{
	COORD Origin		= { 0, 0 };
	DWORD CharsWritten	= 0;


	PalEdit_EraseTopLine();
	Origin.X = 79 - (SHORT) wcslen( Text );
	WriteConsoleOutputCharacter( ConsoleHandle, Text, (DWORD) wcslen( Text ), Origin, &CharsWritten );
	return 0;
}



int PalEdit_ShowFileMsg( LPTSTR Text, LPTSTR Filename )
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";
	wchar_t fn[MAX_FILENAME_LEN]		= L"";


	if ( !_wcsicmp( Filename, DefaultsFilename ) )
		wcscpy_s( fn, MAX_FILENAME_LEN, L"defaults" );
	else {
		_FilenamePart( Filename, fn );

		if ( fn[0] == '\0' )
			wcscpy_s( fn, MAX_FILENAME_LEN, Filename );
	}

	if ( Text )
		wcscpy_s( temp, SHORT_BUF_LEN, Text );

	fn[78 - wcslen( temp )] = '\0';
	wcscat_s( temp, SHORT_BUF_LEN, fn );
	return PalEdit_ShowStringTop( temp );
}



int PalEdit_SaveFile( CONSOLE_SCREEN_BUFFER_INFOEX *BufferInfo, _SaveInfo *SaveInfo )
{
	wchar_t Filename[MAX_FILENAME_LEN]			= L"";
	wchar_t ErrorText[SHORT_BUF_LEN]			= L"";

	OPENFILENAME OpenFilename;
	unsigned int FileType		= PALETTE_FILE_TYPE_NATIVE;
	LPTSTR FileExt				= Filename;
	int rv						= 0;


	memset( &OpenFilename, 0x00, sizeof( OPENFILENAME ) );
	OpenFilename.lStructSize	= sizeof( OPENFILENAME );
	OpenFilename.hwndOwner		= MyWindow();
	OpenFilename.lpstrFilter	= PalEditOpenFilter;
	OpenFilename.lpstrFile		= Filename;
	OpenFilename.nMaxFile		= MAX_FILENAME_LEN;
	OpenFilename.lpstrTitle		= L"Save Palette File:";
	OpenFilename.Flags			= OFN_NOCHANGEDIR | OFN_OVERWRITEPROMPT;


	if ( !GetSaveFileName( &OpenFilename ) )
		return 0;
	else if ( Filename[0] == '\0' )
		return 0;


	if ( OpenFilename.nFileExtension ) {
		FileExt = Filename + OpenFilename.nFileExtension;

		if ( !_wcsicmp( FileExt, L"ini" ) )
			FileType = PALETTE_FILE_TYPE_INI;
		else if ( !_wcsicmp( FileExt, L"xml" ) )
			FileType = PALETTE_FILE_TYPE_XML;
		else if ( !_wcsicmp( FileExt, L"itermcolors" ) )
			FileType = PALETTE_FILE_TYPE_XML;
	}


	SaveInfo->Attribs	= BufferInfo->wAttributes & 0x00ff;

	rv = SaveColorsToFile( Filename, BufferInfo->ColorTable, FileType, PALETTE_FLAG_NO_ERRORS | PALETTE_FLAG_QUIET | PALETTE_FLAG_ALLOW_OVERWRITE, BufferInfo->wAttributes & 0x00ff, SaveInfo );

	if ( rv ) {
		GetErrorText( rv, ErrorText );
		FmtMsgBox( FALSE, PLUGIN_NAME, L"Error saving file:\n%s\n", ErrorText );
		return rv;
	}

	PalEdit_ShowFileMsg( L"File saved:  ", Filename );
	return 0;
}



int PalEdit_LoadFile( CONSOLE_SCREEN_BUFFER_INFOEX *BufferInfo )
{
	wchar_t Filename[MAX_FILENAME_LEN]			= L"";
	wchar_t ErrorText[SHORT_BUF_LEN]			= L"";

	OPENFILENAME OpenFilename;
	COLORREF Colors[16];

	int rv					= 0;


	memset( &OpenFilename, 0x00, sizeof( OPENFILENAME ) );
	OpenFilename.lStructSize	= sizeof( OPENFILENAME );
	OpenFilename.hwndOwner		= MyWindow();
	OpenFilename.lpstrFilter	= PalEditOpenFilter;
	OpenFilename.lpstrFile		= Filename;
	OpenFilename.nMaxFile		= MAX_FILENAME_LEN;
	OpenFilename.lpstrTitle		= L"Load Palette File:";
	OpenFilename.Flags			= OFN_NOCHANGEDIR | OFN_FILEMUSTEXIST;

	memset( Colors, 0x00, 16 * sizeof( COLORREF ) );


	if ( !GetOpenFileName( &OpenFilename ) )
		return 0;
	else if ( Filename[0] == '\0' )
		return 0;

	rv = ReadColorsFromFile( Filename, Colors, PALETTE_FLAG_NO_ERRORS | PALETTE_FLAG_QUIET );

	if ( rv ) {
		GetErrorText( rv, ErrorText );
		FmtMsgBox( FALSE, PLUGIN_NAME, L"Error loading file:\n%s\n", ErrorText );
		return rv;
	}

	PalEdit_UpdatePalette( Colors, -1, BufferInfo );
	PalEdit_ShowFileMsg( L"File loaded:  ", Filename );
	return 0;
}



int PalEdit_DrawScreen( WORD Attribs, CONSOLE_SCREEN_BUFFER_INFOEX *BufferInfo )
{
	COORD Origin			= { 0, 0 };
	DWORD FillLen			= 0;
	DWORD NumDone			= 0;


	FillLen = BufferInfo->dwSize.X * BufferInfo->dwSize.Y;

	Origin.X	= 0;
	Origin.Y	= 0;

	if ( !FillConsoleOutputCharacter( ConsoleHandle, SPACE, FillLen, Origin, &NumDone ) )
		return 2;

	FillConsoleOutputAttribute( ConsoleHandle, Attribs & 0x00ff, FillLen, Origin, &NumDone );
	SetConsoleCursorPosition( ConsoleHandle, Origin );
	SetConsoleTextAttribute( ConsoleHandle, Attribs );

	_cputws( L"\n\n" );

	for ( unsigned int i = 0; i < 16; i++ ) {
		_cwprintf( L"   %X : ", i );
		SetConsoleTextAttribute( ConsoleHandle, ( i << 4 ) | 0x000f );
		_cputws( L"    " );

		for ( unsigned int j = 0; j < 16; j++ ) {
			SetConsoleTextAttribute( ConsoleHandle, ( i << 4 ) | j );
			_cwprintf( L" %02X ", ( i << 4 ) | j );
		}

		_cputws( L"    " );
		SetConsoleTextAttribute( ConsoleHandle, Attribs );
		_cputws( L"\n" );
	}

	_cwprintf( L"\n     0 - F: Edit color    \u2191 \u2193 Background (%X)    \u2190 \u2192 Foreground (%X)    ESC: Exit\n", ( Attribs & 0x00f0 ) >> 4, Attribs & 0x000f );
	_cputws( L"     F5: Refresh    L: Load from file   S: Save to file    ^Z: Undo    ^Y: Redo\n" );

	return 0;
}



int PalEdit_FreeStatesList( _PalEditState *First )
{
	_PalEditState *This		= First;
	_PalEditState *Next		= NULL;

	int k		= 0;


	while ( This ) {
		Next = This->Next;
		free( This );
		This = Next;
		k++;
	}

	return k;
}



DLLExports int WINAPI paledit( LPTSTR inString )
{
	if ( !inString )
		return -1;


	wchar_t Filename[MAX_ARG_LEN]			= L"";

	wchar_t SchemeName[SHORT_BUF_LEN]		= L"";
	wchar_t SchemeAuthor[SHORT_BUF_LEN]		= L"";
	wchar_t SchemeUrl[SHORT_BUF_LEN]		= L"";
	wchar_t SchemeLicense[SHORT_BUF_LEN]	= L"";


	CONSOLE_SCREEN_BUFFER_INFOEX BufferInfo;

	_PalEditState OriginalState;
	_PalEditState *Current		= &OriginalState;
	_PalEditState *New			= NULL;
	unsigned int LatestState	= 0;

	COLORREF NewColor			= 0;

	unsigned int fg				= 0;
	unsigned int bg				= 0;
	unsigned int key			= 0;
	unsigned int index			= 0;

	_SaveInfo SaveInfo;

	BOOL Noisy					= FALSE;
	BOOL NoWarning				= FALSE;

	unsigned int ReadFlags		= PALETTE_FLAG_QUIET;

	int rv						= 0;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( !wcsncmp( arg[i], L"/?", 2 ) )
			return ShowCmdHelp( PalEditHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( PALEDIT_VERSION ) );
		else if ( !_wcsicmp( arg[i], L"/D" ) )
			wcscpy_s( Filename, DefaultsFilename );
		else if ( !_wcsicmp( arg[i], L"/I" ) )
			ReadFlags |= PALETTE_FLAG_ALLOW_INCOMPLETE;
		else if ( towupper( arg[i][0] ) == '/' ) {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
			return 1;
		}
		else if ( Filename[0] == '\0' ) {
			QueryTrueName( arg[i], Filename );

			if ( Filename[0] == '\0' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( L"Bad filename: \"%s\"\r\n", arg[i] );
				return 1;
			}
		}
		else {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
			return 1;
		}
	}


	if ( !NoWarning && !IsWindowVisible( GetConsoleWindow() ) && !ModernTCmdTab() ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"The console window is not visible!\r\n" );
		return 2;
	}


	CtrlBreakFlag = 0;
	SetConsoleCtrlHandler( CtrlBreakHandler, TRUE );


	memset( &BufferInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFOEX ) );
	BufferInfo.cbSize = sizeof( CONSOLE_SCREEN_BUFFER_INFOEX );

	if ( !GetConsoleScreenBufferInfoEx( ConsoleHandle, &BufferInfo ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Unable to get console buffer info\r\n" );
		return 2;
	}


	if ( Filename[0] ) {
		COLORREF Colors[16];

		if ( ReadColorsFromFile( Filename, Colors, ReadFlags ) )
			return 2;

		PalEdit_UpdatePalette( Colors, -1, &BufferInfo );
	}


	fg	= BufferInfo.wAttributes & 0x000f;
	bg	= ( BufferInfo.wAttributes & 0x00f0 ) >> 4;

	memset( &OriginalState, 0x00, sizeof( _PalEditState ) );
	memcpy( OriginalState.Colors, BufferInfo.ColorTable, 16 * sizeof( COLORREF ) );

	SaveInfo.Name		= SchemeName;
	SaveInfo.Author		= SchemeAuthor;
	SaveInfo.Url		= SchemeUrl;
	SaveInfo.License	= SchemeLicense;
	SaveInfo.Attribs	= BufferInfo.wAttributes & 0x00ff;
	SaveInfo.DlgFlags	= PalEdit_GetDlgFlags();


	PalEdit_DrawScreen( ( bg << 4 ) | fg, &BufferInfo );

	if ( Filename[0] ) {
		PalEdit_ShowFileMsg( L"File loaded:  ", Filename );
		Filename[0] = '\0';
	}


	do {
		key = PalEdit_GetKey();

		switch ( key ) {

		case CUR_UP :

			bg = ( bg + 15 ) % 16;
			fg = PalEdit_ContrastyColor( bg, BufferInfo.ColorTable );
			PalEdit_DrawScreen( ( bg << 4 ) | fg, &BufferInfo );
			break;

		case CUR_DOWN :

			bg = ( bg + 1 ) % 16;
			fg = PalEdit_ContrastyColor( bg, BufferInfo.ColorTable );
			PalEdit_DrawScreen( ( bg << 4 ) | fg, &BufferInfo );
			break;

		case CUR_LEFT :

			fg = ( fg + 15 ) % 16;
			if ( fg == bg )
				fg = ( fg + 15 ) % 16;

			PalEdit_DrawScreen( ( bg << 4 ) | fg, &BufferInfo );
			break;

		case CUR_RIGHT :

			fg = ( fg + 1 ) % 16;
			if ( fg == bg )
				fg = ( fg + 1 ) % 16;

			PalEdit_DrawScreen( ( bg << 4 ) | fg, &BufferInfo );
			break;
			
		case 'R' :
		case F5 :

			PalEdit_DrawScreen( ( bg << 4 ) | fg, &BufferInfo );
			RefreshWindows();
			break;

		case F8 :		//	F8 - original state

			if ( Current != &OriginalState ) {
				Current = &OriginalState;
				PalEdit_UpdatePalette( Current->Colors, -1, &BufferInfo );
				PalEdit_ShowUndoInfo( Current->Changes, LatestState );
			}

			break;

		case 0x19 :		//	Ctrl-Y - redo

			if ( Current->Next ) {
				Current = Current->Next;
				PalEdit_UpdatePalette( Current->Colors, -1, &BufferInfo );
				PalEdit_ShowUndoInfo( Current->Changes, LatestState );
			}
			else if ( Noisy )
				honk();

			break;

		case 0x1a :		//	Ctrl-Z - undo

			if ( Current->Prev ) {
				Current = Current->Prev;
				PalEdit_UpdatePalette( Current->Colors, -1, &BufferInfo );
				PalEdit_ShowUndoInfo( Current->Changes, LatestState );
			}
			else if ( Noisy )
				honk();

			break;

		case 'I' :

			DialogBoxParam( MyDLLHandle, MAKEINTRESOURCE( IDD_DIALOG1 ), MyWindow(), PalEdit_SaveInfoDialogProc, (LPARAM) &SaveInfo );
			PalEdit_SaveDlgFlags( SaveInfo.DlgFlags );
			break;

		case 'L' :
		case 0x0c :

			if ( !PalEdit_LoadFile( &BufferInfo ) ) {
				PalEdit_FreeStatesList( Current->Next );
				Current->Next	= NULL;
				LatestState		= Current->Changes;

				New = (_PalEditState*) malloc( sizeof( _PalEditState ) );
				if ( !New ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Out of memory\r\n" );
					rv = 2;
					goto CleanupAndExit;
				}

				Current->Next	= New;
				New->Prev		= Current;
				New->Next		= NULL;
				New->Changes	= Current->Changes + 1;
				memcpy( New->Colors, BufferInfo.ColorTable, 16 * sizeof( COLORREF ) );
				Current			= New;
				LatestState		= Current->Changes;

				SaveInfo.Name[0]	= '\0';
				SaveInfo.Author[0]	= '\0';
				SaveInfo.Url[0]		= '\0';
				SaveInfo.License[0]	= '\0';
			}

			break;

		case 'S' :
		case 0x13 :

			PalEdit_ShowSaveInfoDlgMaybe( &SaveInfo );
			PalEdit_SaveFile( &BufferInfo, &SaveInfo );
			break;

		default :

			if ( ( ( key >= '0' ) && ( key <= '9' ) ) || ( ( key >= 'A' ) && ( key <= 'F' ) ) ) {
				index = iswdigit( key ) ? key - '0' : key - 'A' + 10;
				NewColor = PalEdit_ColorPicker( index, &BufferInfo );

				if ( ( NewColor & 0xff000000 ) == 0 ) {
					PalEdit_FreeStatesList( Current->Next );
					Current->Next	= NULL;
					LatestState		= Current->Changes;

					New = (_PalEditState*) malloc( sizeof( _PalEditState ) );
					if ( !New ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Out of memory\r\n" );
						rv = 2;
						goto CleanupAndExit;
					}

					BufferInfo.ColorTable[index] = NewColor;
					PalEdit_UpdatePalette( &NewColor, index, &BufferInfo );

					Current->Next	= New;
					New->Prev		= Current;
					New->Next		= NULL;
					New->Changes	= Current->Changes + 1;
					memcpy( New->Colors, BufferInfo.ColorTable, 16 * sizeof( COLORREF ) );
					Current			= New;
					LatestState		= Current->Changes;

					PalEdit_EraseTopLine();
				}
			}
		}
	}
	while ( key != ESC );


CleanupAndExit:

	SetConsoleCtrlHandler( CtrlBreakHandler, FALSE );
	PalEdit_FreeStatesList( OriginalState.Next );
	PalEdit_UpdateStdColors( bg, fg );
	return rv;
}


#endif	//	PALEDIT_VERSION
